@@ -23,12 +23,11 @@
 
 /* includes */
 #include "pmacct.h"
+#include "nfacctd.h"
 #include "pmacct-data.h"
 #include "plugin_hooks.h"
 #include "cfg_handlers.h"
 
-#define OFF(x, y) ((unsigned long int)y - (unsigned long int)x)
-
 void lower_string(char *string)
 {
   int i = 0;
@@ -39,128 +38,6 @@ void lower_string(char *string)
   }
 }
 
-int set_all_int(unsigned int off, int value)
-{
-  struct plugins_list_entry *list = plugins_list;
-  unsigned int **ptr;
-  int changes = 0;
-
-  while (list) {
-    ptr = (unsigned int **) &list->cfg;
-
-#ifdef OSF1
-    if (off%sizeof(unsigned int **)) (u_char) ptr += off;
-    else ptr += off/sizeof(unsigned int **);
-#else
-    ptr += off/sizeof(unsigned int **);
-#endif
-    
-#ifdef IRIX
-    *ptr = (unsigned int *) value;
-#else
-    (int) *ptr = value;
-#endif
-
-    changes++;
-    list = list->next;
-  }
-
-  return changes;
-}
-
-int set_one_int(unsigned int off, char *name, int value)
-{
-  struct plugins_list_entry *list = plugins_list;
-  unsigned int **ptr;
-  int changes = 0;
-
-  while (list) {
-    if (!strcmp(name, list->name)) {
-      ptr = (unsigned int **) &list->cfg;
-
-#ifdef OSF1
-      if (off%sizeof(unsigned int **)) (u_char) ptr += off;
-      else ptr += off/sizeof(unsigned int **);
-#else
-      ptr += off/sizeof(unsigned int **);
-#endif
-
-#ifdef IRIX
-      *ptr = (unsigned int *) value;
-#else
-      (int) *ptr = value;
-#endif
-
-      changes++;
-      break;
-    }
-    list = list->next;
-  }
-
-  return changes;
-}
-
-int set_all_char_ptr(unsigned int off, char *value)
-{
-  struct plugins_list_entry *list = plugins_list;
-  unsigned int **ptr;
-  int changes = 0;
-
-  while (list) {
-    ptr = (unsigned int **) &list->cfg;
-
-#ifdef OSF1
-    if (off%sizeof(unsigned int **)) (u_char) ptr += off;
-    else ptr += off/sizeof(unsigned int **);
-#else
-    ptr += off/sizeof(unsigned int **);
-#endif
-
-#ifdef IRIX
-    *ptr = (unsigned int *) value;
-#else
-    (char *) *ptr = value;
-#endif
-
-    changes++;
-    list = list->next;
-  }
-
-  return changes;
-}
-
-int set_one_char_ptr(unsigned int off, char *name, char *value)
-{
-  struct plugins_list_entry *list = plugins_list;
-  unsigned int **ptr;
-  int changes = 0;
-
-  while (list) {
-    if (!strcmp(name, list->name)) {
-      ptr = (unsigned int **) &list->cfg;
-
-#ifdef OSF1
-      if (off%sizeof(unsigned int **)) (u_char) ptr += off;
-      else ptr += off/sizeof(unsigned int **);
-#else
-      ptr += off/sizeof(unsigned int **);
-#endif
-
-#ifdef IRIX
-      *ptr = (unsigned int *) value;
-#else
-      (char *) *ptr = value;
-#endif
-
-      changes++;
-      break;
-    }
-    list = list->next;
-  }
-
-  return changes;
-}
-
 int parse_truefalse(char *value_ptr)
 {
   int value, i = 0;
@@ -182,8 +59,16 @@ int cfg_key_debug(char *name, char *value_ptr)
   value = parse_truefalse(value_ptr); 
   if (value < 0) return ERR; 
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.debug), value);
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.debug), name, value);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.debug = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) { 
+	list->cfg.debug = value;
+	changes++;
+	break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -193,8 +78,16 @@ int cfg_key_syslog(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.syslog), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.syslog), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.syslog = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.syslog = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -204,7 +97,7 @@ int cfg_key_pidfile(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.pidfile), value_ptr);
+  for (; list; list = list->next, changes++) list->cfg.pidfile = value_ptr;
   if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'pidfile'. Globalized.\n");
 
   return changes;
@@ -218,7 +111,7 @@ int cfg_key_daemonize(char *name, char *value_ptr)
   value = parse_truefalse(value_ptr);
   if (value < 0) return ERR;
 
-  changes = set_all_int(OFF(&list->cfg, &list->cfg.daemon), value);
+  for (; list; list = list->next, changes++) list->cfg.daemon = value;
   if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'daemonize'. Globalized.\n"); 
 
   return changes;
@@ -244,11 +137,20 @@ int cfg_key_aggregate(char *name, char *value_ptr)
     else if (!strcmp(count_token, "proto")) value |= COUNT_IP_PROTO;
     else if (!strcmp(count_token, "src_mac")) value |= COUNT_SRC_MAC;
     else if (!strcmp(count_token, "dst_mac")) value |= COUNT_DST_MAC;
+    else if (!strcmp(count_token, "vlan")) value |= COUNT_VLAN;
     else Log(LOG_WARNING, "WARN: ignoring unknown aggregation method: %s.\n", count_token);
   }
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.what_to_count), value);
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.what_to_count), name, value);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.what_to_count = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.what_to_count = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -262,17 +164,61 @@ int cfg_key_aggregate_filter(char *name, char *value_ptr)
     Log(LOG_ERR, "ERROR: aggregation filter cannot be global. Not loaded.\n");
     changes++;
   }
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.a_filter), name, value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.a_filter = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
 
+int cfg_key_pre_tag_filter(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  if (config.acct_type != ACCT_NF) {
+    Log(LOG_ERR, "ERROR: Pre Tag filter is not applicable to current data source. Not loaded.\n");
+    changes++;
+    return changes;
+  } 
+
+  value = atoi(value_ptr);
+  if ((value < 1) || (value > 65535)) {
+    Log(LOG_ERR, "WARN: 'pre_tag_filter' has to be in the range 1-65535. Not loaded.\n");
+    changes++;
+    return changes;
+  }
+
+  if (!name) {
+    Log(LOG_ERR, "ERROR: ID filter cannot be global. Not loaded.\n");
+    changes++;
+  }
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.pre_tag_filter = value;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+
 int cfg_key_pcap_filter(char *name, char *value_ptr)
 {
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.clbuf), value_ptr);
+  for (; list; list = list->next, changes++) list->cfg.clbuf = value_ptr;
   if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'pcap_filter'. Globalized.\n");
 
   return changes;
@@ -283,7 +229,7 @@ int cfg_key_interface(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.dev), value_ptr);
+  for (; list; list = list->next, changes++) list->cfg.dev = value_ptr;
   if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'interface'. Globalized.\n");
 
   return changes;
@@ -297,7 +243,7 @@ int cfg_key_promisc(char *name, char *value_ptr)
   value = parse_truefalse(value_ptr);
   if (value < 0) return ERR;
 
-  changes = set_all_int(OFF(&list->cfg, &list->cfg.promisc), value);
+  for (; list; list = list->next, changes++) list->cfg.promisc = value;
   if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'promisc'. Globalized.\n");
 
   return changes;
@@ -308,8 +254,16 @@ int cfg_key_imt_path(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.imt_plugin_path), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.imt_plugin_path), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.imt_plugin_path = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.imt_plugin_path = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -319,8 +273,16 @@ int cfg_key_imt_passwd(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.imt_plugin_passwd), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.imt_plugin_passwd), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.imt_plugin_passwd = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.imt_plugin_passwd = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -336,8 +298,16 @@ int cfg_key_imt_buckets(char *name, char *value_ptr)
     exit(1);
   }
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.buckets), value);
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.buckets), name, value);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.buckets = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.buckets = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -353,8 +323,16 @@ int cfg_key_imt_mem_pools_number(char *name, char *value_ptr)
     exit(1);
   }
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.num_memory_pools), value);
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.num_memory_pools), name, value);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.num_memory_pools = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.num_memory_pools = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   have_num_memory_pools = TRUE;
   return changes;
@@ -365,8 +343,16 @@ int cfg_key_imt_mem_pools_size(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.memory_pool_size), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.memory_pool_size), name, atoi(value_ptr));
+  if (!name) for (; list; list = list->next, changes++) list->cfg.memory_pool_size = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.memory_pool_size = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -376,8 +362,16 @@ int cfg_key_sql_db(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_db), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_db), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_db = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_db = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -387,8 +381,41 @@ int cfg_key_sql_table(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_table), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_table), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_table = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_table = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_sql_table_version(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = atoi(value_ptr);
+  if (value <= 0) {
+    Log(LOG_ERR, "ERROR: invalid 'sql_table_version' value.\n");
+    exit(1);
+  }
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_table_version = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_table_version = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -398,8 +425,16 @@ int cfg_key_sql_data(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_data), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_data), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_data = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_data = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -409,8 +444,16 @@ int cfg_key_sql_host(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_host), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_host), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_host = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_host = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -420,8 +463,62 @@ int cfg_key_sql_recovery_backup_host(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_backup_host), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_backup_host), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_backup_host = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_backup_host = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_sql_trigger_exec(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_trigger_exec = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_trigger_exec = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_sql_trigger_time(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0, t, t_howmany;
+
+  parse_time(value_ptr, &t, &t_howmany);
+
+  if (!name) {
+    for (; list; list = list->next, changes++) {
+      list->cfg.sql_trigger_time = t;
+      list->cfg.sql_trigger_time_howmany = t_howmany;
+    }
+  }
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_trigger_time = t;
+	list->cfg.sql_trigger_time_howmany = t_howmany;
+	changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -431,8 +528,16 @@ int cfg_key_sql_user(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_user), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_user), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_user = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_user = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -442,8 +547,16 @@ int cfg_key_sql_passwd(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_passwd), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_passwd), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_passwd = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_passwd = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -453,8 +566,22 @@ int cfg_key_sql_refresh_time(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.sql_refresh_time), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.sql_refresh_time), name, atoi(value_ptr));
+  value = atoi(value_ptr);
+  if (value <= 0) {
+    Log(LOG_ERR, "WARN: 'sql_refresh_time' has to be > 0\n");
+    return ERR;
+  }
+     
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_refresh_time = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_refresh_time = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -464,8 +591,16 @@ int cfg_key_sql_startup_delay(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.sql_startup_delay), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.sql_startup_delay), name, atoi(value_ptr));
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_startup_delay = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_startup_delay = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -478,8 +613,16 @@ int cfg_key_sql_optimize_clauses(char *name, char *value_ptr)
   value = parse_truefalse(value_ptr);
   if (value < 0) return ERR;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.sql_optimize_clauses), value);
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.sql_optimize_clauses), name, value);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_optimize_clauses = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_optimize_clauses = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -489,8 +632,16 @@ int cfg_key_sql_history_roundoff(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_history_roundoff), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_history_roundoff), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_history_roundoff = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_history_roundoff = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -500,8 +651,16 @@ int cfg_key_sql_recovery_logfile(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.sql_recovery_logfile), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.sql_recovery_logfile), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_recovery_logfile = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_recovery_logfile = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -511,15 +670,23 @@ int cfg_key_sql_history(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0, sql_history, sql_history_howmany;
 
-  parse_history_time(value_ptr, &sql_history, &sql_history_howmany);
+  parse_time(value_ptr, &sql_history, &sql_history_howmany);
 
   if (!name) {
-    changes = set_all_int(OFF(&list->cfg, &list->cfg.sql_history), sql_history);
-    changes += set_all_int(OFF(&list->cfg, &list->cfg.sql_history_howmany), sql_history_howmany);
+    for (; list; list = list->next, changes++) {
+      list->cfg.sql_history = sql_history;
+      list->cfg.sql_history_howmany = sql_history_howmany;
+    }
   }
   else {
-    changes = set_one_int(OFF(&list->cfg, &list->cfg.sql_history), name, sql_history);
-    changes += set_one_int(OFF(&list->cfg, &list->cfg.sql_history_howmany), name, sql_history_howmany);
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_history = sql_history;
+        list->cfg.sql_history_howmany = sql_history_howmany;
+        changes++;
+        break;
+      }
+    }
   }
 
   return changes;
@@ -530,8 +697,38 @@ int cfg_key_sql_cache_entries(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.sql_cache_entries), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.sql_cache_entries), name, atoi(value_ptr));
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_cache_entries = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_cache_entries = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_sql_dont_try_update(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = parse_truefalse(value_ptr);
+  if (value < 0) return ERR;
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sql_dont_try_update = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sql_dont_try_update = value;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -541,8 +738,16 @@ int cfg_key_plugin_pipe_size(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.pipe_size), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.pipe_size), name, atoi(value_ptr));
+  if (!name) for (; list; list = list->next, changes++) list->cfg.pipe_size = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.pipe_size = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -552,8 +757,16 @@ int cfg_key_plugin_buffer_size(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_int(OFF(&list->cfg, &list->cfg.buffer_size), atoi(value_ptr));
-  else changes = set_one_int(OFF(&list->cfg, &list->cfg.buffer_size), name, atoi(value_ptr));
+  if (!name) for (; list; list = list->next, changes++) list->cfg.buffer_size = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.buffer_size = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
@@ -563,13 +776,215 @@ int cfg_key_networks_file(char *name, char *value_ptr)
   struct plugins_list_entry *list = plugins_list;
   int value, changes = 0;
 
-  if (!name) changes = set_all_char_ptr(OFF(&list->cfg, &list->cfg.networks_file), value_ptr);
-  else changes = set_one_char_ptr(OFF(&list->cfg, &list->cfg.networks_file), name, value_ptr);
+  if (!name) for (; list; list = list->next, changes++) list->cfg.networks_file = value_ptr;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.networks_file = value_ptr;
+        changes++;
+        break;
+      }
+    }
+  }
 
   return changes;
 }
 
-void parse_history_time(char *value, int *mu, int *howmany)
+int cfg_key_print_refresh_time(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = atoi(value_ptr);
+  if (value <= 0) {
+    Log(LOG_ERR, "WARN: 'print_refresh_time' has to be > 0\n");
+    return ERR;
+  }
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.print_refresh_time = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.print_refresh_time = value;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_print_cache_entries(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.print_cache_entries = atoi(value_ptr);
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.print_cache_entries = atoi(value_ptr);
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_print_markers(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = parse_truefalse(value_ptr);
+  if (value < 0) return ERR;
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.print_markers = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.print_markers = value;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_post_tag(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = atoi(value_ptr);
+  if ((value < 1) || (value > 65535)) {
+    Log(LOG_ERR, "WARN: 'post_tag' has to be in the range 1-65535.\n");
+    return ERR;
+  }
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.post_tag = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.post_tag = value;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_sampling_rate(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = atoi(value_ptr);
+  if (value < 1) {
+    Log(LOG_ERR, "WARN: 'sampling_rate' has to be >= 1.\n");
+    return ERR;
+  }
+
+  if (!name) for (; list; list = list->next, changes++) list->cfg.sampling_rate = value;
+  else {
+    for (; list; list = list->next) {
+      if (!strcmp(name, list->name)) {
+        list->cfg.sampling_rate = value;
+        changes++;
+        break;
+      }
+    }
+  }
+
+  return changes;
+}
+
+int cfg_key_nfacctd_port(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = atoi(value_ptr);
+  if ((value <= 0) || (value > 65535)) {
+    Log(LOG_ERR, "WARN: 'nfacctd_port' has to be in the range 0-65535.\n");
+    return ERR;
+  }
+
+  for (; list; list = list->next, changes++) list->cfg.nfacctd_port = value;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'nfacctd_port'. Globalized.\n");
+
+  return changes;
+}
+
+int cfg_key_nfacctd_ip(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  for (; list; list = list->next, changes++) list->cfg.nfacctd_ip = value_ptr;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'nfacctd_ip'. Globalized.\n");
+
+  return changes;
+}
+
+int cfg_key_nfacctd_allow_file(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  for (; list; list = list->next, changes++) list->cfg.nfacctd_allow_file = value_ptr;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'nfacctd_allow_file'. Globalized.\n");
+
+  return changes;
+}
+
+int cfg_key_pre_tag_map(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  for (; list; list = list->next, changes++) list->cfg.pre_tag_map = value_ptr;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'pre_tag_map'. Globalized.\n");
+
+  return changes;
+}
+
+int cfg_key_nfacctd_time_secs(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = parse_truefalse(value_ptr);
+  if (value < 0) return ERR;
+
+  for (; list; list = list->next, changes++) list->cfg.nfacctd_time = NF_TIME_SECS;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'nfacctd_time_secs'. Globalized.\n");
+
+  return changes;
+}
+
+int cfg_key_nfacctd_time_new(char *name, char *value_ptr)
+{
+  struct plugins_list_entry *list = plugins_list;
+  int value, changes = 0;
+
+  value = parse_truefalse(value_ptr);
+  if (value < 0) return ERR;
+
+  for (; list; list = list->next, changes++) list->cfg.nfacctd_time = NF_TIME_NEW;
+  if (name) Log(LOG_WARNING, "WARN: plugin name not supported for key 'nfacctd_time_new'. Globalized.\n");
+
+  return changes;
+}
+
+void parse_time(char *value, int *mu, int *howmany)
 {
   int k, j, len;
 
