@@ -34,7 +34,7 @@
 #include "crc32.c"
 
 /* Functions */
-void mysql_plugin(int pipe_fd, struct configuration *cfgptr) 
+void mysql_plugin(int pipe_fd, struct configuration *cfgptr, void *ptr) 
 {
   struct pkt_data *data;
   struct networks_table nt;
@@ -43,11 +43,21 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
   char mysql_pwd[] = "arealsmartpwd";
   char mysql_db[] = "pmacct"; 
   char mysql_table[] = "acct";
+  char mysql_table_v2[] = "acct_v2";
   struct pollfd pfd;
   struct insert_data idata;
   time_t t, now, refresh_deadline;
-  int timeout = config.sql_refresh_time*1000; /* dirty */ 
+  int timeout;
   int ret;
+#if defined (HAVE_MMAP)
+  struct ring *rg = &((struct channels_list_entry *)ptr)->rg;
+  struct ch_status *status = ((struct channels_list_entry *)ptr)->status;
+  u_int32_t bufsz = ((struct channels_list_entry *)ptr)->bufsize;
+
+  unsigned char *rgptr;
+  int pollagain = TRUE;
+  u_int32_t seq = 1, rg_err_count = 0; 
+#endif
 
   /* XXX: glue */
   memcpy(&config, cfgptr, sizeof(struct configuration));
@@ -59,12 +69,17 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
   /* checks */
   if (config.what_to_count & COUNT_SUM_HOST) {
     Log(LOG_ERR, "ERROR: Option available only in memory table operations\nExiting ...\n\n");
-    exit(0);
+    exit(1);
   }
 
   if (!config.sql_refresh_time)
     config.sql_refresh_time = DEFAULT_DB_REFRESH_TIME;
 
+  if (!config.sql_table_version)
+    config.sql_table_version = DEFAULT_SQL_TABLE_VERSION;
+
+  timeout = config.sql_refresh_time*1000; /* dirty */
+
   if (config.networks_file) load_networks(config.networks_file, &nt);
   
   memset(sql_data, 0, sizeof(sql_data));
@@ -95,7 +110,10 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
   if (!config.sql_user) config.sql_user = mysql_user; 
   if (!config.sql_db) config.sql_db = mysql_db;
   if (!config.sql_passwd) config.sql_passwd = mysql_pwd; 
-  if (!config.sql_table) config.sql_table = mysql_table;
+  if (!config.sql_table) {
+    if (config.sql_table_version == 2) config.sql_table = mysql_table_v2;
+    else config.sql_table = mysql_table;
+  }
 
   now = time(NULL);
 
@@ -115,6 +133,28 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
     idata.basetime = t;
   }
 
+  /* sql triggers time init: deadline; if a trigger exec is specified but no
+     time is supplied, use 'sql_refresh_time' as interval; this will result in
+     a trigger being executed each time data is purged into the DB */
+  if (config.sql_trigger_exec) {
+    time_t deadline;
+
+    deadline = now;
+    if (config.sql_trigger_time == COUNT_MINUTELY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*60;
+    else if (config.sql_trigger_time == COUNT_HOURLY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*3600;
+    else if (config.sql_trigger_time == COUNT_DAILY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*86400;
+    else config.sql_trigger_time_howmany = config.sql_refresh_time; 
+
+    /* round off stuff */
+    t = roundoff_time(deadline);
+    while ((t+config.sql_trigger_time_howmany) < deadline) t += config.sql_trigger_time_howmany;
+    config.sql_trigger_time = t;
+    config.sql_trigger_time += config.sql_trigger_time_howmany; /* it's a deadline not a basetime */
+  }
+
   /* sql_refresh time init: deadline */
   refresh_deadline = now; 
   t = roundoff_time(refresh_deadline);
@@ -136,8 +176,14 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
   memset(collision_queue, 0, cq_size*sizeof(struct db_cache *));
 
   /* plugin main loop */
-  while(1) {
+  for(;;) {
+    poll_again:
+#if defined (HAVE_MMAP)
+    status->wakeup = TRUE;
+#endif
     ret = poll(&pfd, 1, timeout);
+    if (ret < 0) goto poll_again;
+
     now = time(NULL);
 
     switch (ret) {
@@ -154,8 +200,11 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
 	  memset(&b, 0, sizeof(b));
 
           if (!MY_DB_Connect(&p, config.sql_host)) Log(LOG_ALERT, "ALERT: MySQL daemon failed.\n");
-          if (cq_ptr) MY_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
           MY_cache_purge(queries_queue, qq_ptr, idata.num_primitives, TRUE);
+          if (cq_ptr) {
+	    config.sql_dont_try_update = FALSE;
+	    MY_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
+	  }
           
 	  if (!p.fail) mysql_close(&p.desc);
   	  else if (b.connected) mysql_close(&b.desc);
@@ -169,8 +218,47 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
       }
       break;
     default: /* we received data */
-      if (read(pipe_fd, pipebuf, config.buffer_size) == 0) 
-        exit(0); /* we exit silently; something happened at the write end */
+#if !defined (HAVE_MMAP)
+      if ((ret = read(pipe_fd, pipebuf, config.buffer_size)) == 0) 
+        exit(1); /* we exit silently; something happened at the write end */
+
+      if (ret < 0) goto poll_again;
+#else
+      read_data:
+      if (!pollagain) {
+        seq++;
+        seq %= MAX_SEQNUM;
+	if (seq == 0) rg_err_count = FALSE;
+	now = time(NULL); 
+      }
+      else {
+        if ((ret = read(pipe_fd, &rgptr, sizeof(rgptr))) == 0) 
+	  exit(1); /* we exit silently; something happened at the write end */
+      }
+
+      if (((struct ch_buf_hdr *)rg->ptr)->seq != seq) {
+	if (!pollagain) {
+	  pollagain = TRUE;
+	  goto poll_again;
+        }
+	else {
+	  rg_err_count++;
+	  if (config.debug || (rg_err_count > MAX_RG_COUNT_ERR)) {
+            Log(LOG_ERR, "ERROR: We are missing data.\n");
+            Log(LOG_ERR, "If you see this message once in a while, discard it. Otherwise some solutions follow:\n");
+            Log(LOG_ERR, "- increase shared memory size, 'plugin_pipe_size'; now: '%d'.\n", config.pipe_size);
+            Log(LOG_ERR, "- increase buffer size, 'plugin_buffer_size'; now: '%d'.\n", config.buffer_size);
+            Log(LOG_ERR, "- increase system maximum socket size.\n\n");
+	  }
+          seq = ((struct ch_buf_hdr *)rg->ptr)->seq;
+	}
+      }
+
+      pollagain = FALSE;
+      memcpy(pipebuf, rg->ptr, bufsz);
+      if ((rg->ptr+bufsz) >= rg->end) rg->ptr = rg->base;
+      else rg->ptr += bufsz;
+#endif
 
       /* lazy sql refresh handling */ 
       if (now > refresh_deadline) {
@@ -186,8 +274,15 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
             memset(&b, 0, sizeof(b));
 
             if (!MY_DB_Connect(&p, config.sql_host)) Log(LOG_ALERT, "ALERT: MySQL daemon failed.\n");
-            if (cq_ptr) MY_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
             MY_cache_purge(queries_queue, qq_ptr, idata.num_primitives, TRUE);
+            if (cq_ptr) {
+	      config.sql_dont_try_update = FALSE;
+	      MY_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
+	    }
+
+	    if (config.sql_trigger_exec) {
+	      if (now > config.sql_trigger_time) MY_Exec(config.sql_trigger_exec);
+	    }
 
             if (!p.fail) mysql_close(&p.desc);
             else if (b.connected) mysql_close(&b.desc);
@@ -196,10 +291,19 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
 	    if (cq_ptr) cq_ptr = MY_cache_flush(&collision_queue, cq_ptr, FALSE);
             qq_ptr = MY_cache_flush(queries_queue, qq_ptr, TRUE);
 	    refresh_deadline += config.sql_refresh_time; 
+            if (now > config.sql_trigger_time) config.sql_trigger_time  += config.sql_trigger_time_howmany;
             break;
 	  }
         }
       } 
+      else {
+	if (config.sql_trigger_exec) {
+	  if (now > config.sql_trigger_time) {
+	    MY_Exec(config.sql_trigger_exec); 
+	    config.sql_trigger_time  += config.sql_trigger_time_howmany;
+	  }
+	}
+      }
 
       data = (struct pkt_data *) (pipebuf+sizeof(struct ch_buf_hdr));
       if (now > (idata.basetime + idata.timeslot)) idata.basetime += idata.timeslot;
@@ -212,10 +316,13 @@ void mysql_plugin(int pipe_fd, struct configuration *cfgptr)
 	  binsearch(&nt, &data->primitives.dst_ip);
 
         idata.modulo = MY_cache_modulo(&data->primitives); 
-        MY_cache_insert(&data->primitives, data->pkt_len, &idata); 
+        MY_cache_insert(data, &idata); 
 	((struct ch_buf_hdr *)pipebuf)->num--;
         if (((struct ch_buf_hdr *)pipebuf)->num) data++;
       }
+#if defined (HAVE_MMAP)
+      goto read_data;
+#endif
     }
   }
 }
@@ -224,7 +331,7 @@ unsigned int MY_cache_modulo(struct pkt_primitives *srcdst)
 {
   register unsigned int modulo;
 
-  modulo = cache_crc32((char *)srcdst, pp_size);
+  modulo = cache_crc32((unsigned char *)srcdst, pp_size);
   
   return modulo %= config.sql_cache_entries;
 }
@@ -234,26 +341,42 @@ int MY_cache_dbop(MYSQL *db_desc, const struct db_cache *cache_elem, const int n
   char *ptr_values, *ptr_where;
   int num=0, ret=0;
 
+  /* constructing sql query */
   ptr_where = where_clause;
   ptr_values = values_clause; 
+  memset(where_clause, 0, sizeof(where_clause));
+  memset(values_clause, 0, sizeof(values_clause));
   while (num < num_primitives) {
     (*where[num].handler)(cache_elem, num, &ptr_values, &ptr_where);
     num++;
   }
   
-  /* constructing sql query */
   snprintf(sql_data, sizeof(sql_data), update_clause, cache_elem->packet_counter, cache_elem->bytes_counter);
   strncat(sql_data, where_clause, SPACELEFT(sql_data));
-  ret = mysql_query(db_desc, sql_data);
-  if (ret) return ret; 
 
-  if (mysql_affected_rows(db_desc) == 0) {
+  /* sending UPDATE query */
+  if (!config.sql_dont_try_update) {
+    ret = mysql_query(db_desc, sql_data);
+    if (ret) {
+      Log(LOG_DEBUG, "FAILED query follows:\n%s\n", sql_data);
+      Log(LOG_ERR, "ERROR: %s\n\n", mysql_error(db_desc));
+      if (mysql_errno(db_desc) == 1062) return FALSE; /* not signalling duplicate entry problems */ 
+      else return ret; 
+    }
+  }
+
+  if (config.sql_dont_try_update || (mysql_affected_rows(db_desc) == 0)) {
     /* UPDATE failed, trying with an INSERT query */ 
     strncpy(sql_data, insert_clause, sizeof(sql_data));
     snprintf(ptr_values, SPACELEFT(values_clause), ", %d, %lu)", cache_elem->packet_counter, cache_elem->bytes_counter);
     strncat(sql_data, values_clause, SPACELEFT(sql_data));
     ret = mysql_query(db_desc, sql_data);
-    if (ret) return ret;
+    if (ret) {
+      Log(LOG_DEBUG, "FAILED query follows:\n%s\n", sql_data);
+      Log(LOG_ERR, "ERROR: %s\n\n", mysql_error(db_desc));
+      if (mysql_errno(db_desc) == 1062) return FALSE; /* not signalling duplicate entry problems */
+      else return ret;
+    }
   }
 
   if (config.debug) Log(LOG_DEBUG, "%s\n\n", sql_data);
@@ -261,10 +384,14 @@ int MY_cache_dbop(MYSQL *db_desc, const struct db_cache *cache_elem, const int n
   return ret;
 }
 
-void MY_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct insert_data *idata) 
+void MY_cache_insert(struct pkt_data *data, struct insert_data *idata) 
 {
   int chances = 2;
   unsigned int modulo = idata->modulo;
+  unsigned long int basetime = idata->basetime;
+  struct pkt_primitives *srcdst = &data->primitives;
+
+  while ((basetime > data->pkt_time) && (data->pkt_time > 0)) basetime -= idata->timeslot;
 
   start:
   if (memcmp(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives)) != 0) { 
@@ -284,29 +411,15 @@ void MY_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct
 	  modulo--;
 	  modulo %= config.sql_cache_entries;
           idata->modulo = modulo;
-          chances--;
 	}
-	else idata->modulo = config.sql_cache_entries;
+	else idata->modulo = config.sql_cache_entries-1;
+        chances--;
 
         goto start;
       }
 
       if (config.debug) Log(LOG_DEBUG, "*** Entry aliasing ***\n");
-      if (cq_ptr < cq_size) {
-        memcpy(&collision_queue[cq_ptr], &cache[idata->modulo], dbc_size);
-        cq_ptr++;
-      }
-      else { 
-	/* purging collision queue */
-        switch (fork()) {
-        case 0: /* Child */
-          MY_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
-          exit(0);
-        default: /* Parent */
-          cq_ptr = MY_cache_flush(&collision_queue, cq_ptr, FALSE);
-          break;
-        }
-      }
+      MY_handle_collision(&cache[idata->modulo]);
     }
     else {
       queries_queue[qq_ptr] = &cache[idata->modulo];
@@ -315,47 +428,41 @@ void MY_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct
 
     /* we add the new entry in the cache */
     memcpy(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives));
-    cache[idata->modulo].packet_counter = 1;
-    cache[idata->modulo].bytes_counter = ntohs(ip_len);
+    cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+    cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
     cache[idata->modulo].valid = TRUE;
-    cache[idata->modulo].basetime = idata->basetime;
+    cache[idata->modulo].basetime = basetime;
   }
   else {
     if (cache[idata->modulo].valid == TRUE) {
       /* additional check: time */
-      if ((cache[idata->modulo].basetime < idata->basetime) && (config.sql_history)) {
+      if ((cache[idata->modulo].basetime < basetime) && (config.sql_history)) {
 	if (config.debug) Log(LOG_DEBUG, "*** Entry aliasing ***\n");
-	if (cq_ptr < cq_size) {
-          memcpy(&collision_queue[cq_ptr], &cache[idata->modulo], dbc_size);
-          cq_ptr++;
-	}
-        else {
-          /* purging collision queue */
-          switch (fork()) {
-          case 0: /* Child */
-            MY_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
-	    exit(0);
-          default: /* Parent */
-            cq_ptr = MY_cache_flush(&collision_queue, cq_ptr, FALSE);
-            break;
-          }
-	}
-	cache[idata->modulo].packet_counter = 1;
-	cache[idata->modulo].bytes_counter = ntohs(ip_len);
-	cache[idata->modulo].basetime = idata->basetime;
+	MY_handle_collision(&cache[idata->modulo]);
+	cache[idata->modulo].packet_counter = ntohl(data->pkt_num); 
+	cache[idata->modulo].bytes_counter =  ntohl(data->pkt_len);
+	cache[idata->modulo].basetime = basetime;
+      }
+      /* additional check: would counters overflow ? */
+      else if ((cache[idata->modulo].packet_counter > UINT32TMAX) ||
+	       (cache[idata->modulo].bytes_counter > UINT32TMAX)) {
+	MY_handle_collision(&cache[idata->modulo]);
+        cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+        cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
+	cache[idata->modulo].basetime = basetime;
       }
       else {
         /* everything is ok; summing counters */
-        cache[idata->modulo].packet_counter++;
-        cache[idata->modulo].bytes_counter += ntohs(ip_len);
+        cache[idata->modulo].packet_counter += ntohl(data->pkt_num);
+        cache[idata->modulo].bytes_counter += ntohl(data->pkt_len);
       }
     }
     else {
-      /* entry invalidated by SIGALRM; restarting counters */
-      cache[idata->modulo].packet_counter = 1;
-      cache[idata->modulo].bytes_counter = ntohs(ip_len);
+      /* entry invalidated; restarting counters */
+      cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+      cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
       cache[idata->modulo].valid = TRUE;
-      cache[idata->modulo].basetime = idata->basetime;
+      cache[idata->modulo].basetime = basetime;
       queries_queue[qq_ptr] = &cache[idata->modulo];
       qq_ptr++;
     }
@@ -405,6 +512,29 @@ int MY_cache_flush(struct db_cache *queue[], int index, int ptr)
   return index;
 }
 
+void MY_handle_collision(struct db_cache *elem)
+{
+  if (cq_ptr < cq_size) {
+    memcpy(&collision_queue[cq_ptr], elem, dbc_size);
+    cq_ptr++;
+  }
+  else {
+    /* purging collision queue */
+    switch (fork()) {
+    case 0: /* Child */
+      config.sql_dont_try_update = FALSE;
+      MY_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
+      exit(0);
+    default: /* Parent */
+      cq_ptr = MY_cache_flush(&collision_queue, cq_ptr, FALSE);
+      break;
+    }
+
+    memcpy(&collision_queue[cq_ptr], elem, dbc_size);
+    cq_ptr++;
+  }
+}
+
 int MY_evaluate_history(int primitive)
 {
   if (config.sql_history) {
@@ -430,23 +560,28 @@ int MY_evaluate_history(int primitive)
 int MY_evaluate_primitives(int primitive)
 {
   register unsigned long int what_to_count=0;
+  register unsigned long int fakes=0;
+  short int assume_custom_table = FALSE; 
 
-  if (config.sql_optimize_clauses) what_to_count = config.what_to_count;
+  if (config.sql_optimize_clauses) {
+    what_to_count = config.what_to_count;
+    assume_custom_table = TRUE;
+  }
   else {
     /* we are requested to avoid optimization;
        then we'll construct an all-true "what
        to count" bitmap */ 
     if (config.what_to_count & COUNT_SRC_MAC) what_to_count |= COUNT_SRC_MAC;
-    else what_to_count |= FAKE_SRC_MAC;
+    else fakes |= FAKE_SRC_MAC;
     if (config.what_to_count & COUNT_DST_MAC) what_to_count |= COUNT_DST_MAC;
-    else what_to_count |= FAKE_DST_MAC;
+    else fakes |= FAKE_DST_MAC;
 
     if (config.what_to_count & COUNT_SRC_HOST) what_to_count |= COUNT_SRC_HOST;
-    else what_to_count |= FAKE_SRC_HOST;
+    else fakes |= FAKE_SRC_HOST;
     if (config.what_to_count & COUNT_DST_HOST) what_to_count |= COUNT_DST_HOST;
-    else what_to_count |= FAKE_DST_HOST;
+    else fakes |= FAKE_DST_HOST;
 
-    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO;
+    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO|COUNT_ID|COUNT_VLAN;
   }
 
   /* 1st part: arranging pointers to an opaque structure and 
@@ -480,6 +615,33 @@ int MY_evaluate_primitives(int primitive)
     primitive++;
   }
 
+  if (what_to_count & COUNT_VLAN) {
+    int count_it = FALSE;
+
+    if ((config.sql_table_version < 2) && !assume_custom_table) {
+      if (config.what_to_count & COUNT_VLAN) {
+        Log(LOG_ERR, "ERROR: The use of VLAN accounting requires SQL table v2. Exiting.\n");
+        exit(1);
+      }
+      else what_to_count ^= COUNT_VLAN;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "vlan", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "vlan=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_VLAN;
+      values[primitive].handler = where[primitive].handler = count_vlan_handler;
+      primitive++;
+    }
+  }
+
   if (what_to_count & COUNT_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
@@ -550,7 +712,34 @@ int MY_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_MAC) {
+  if (what_to_count & COUNT_ID) {
+    int count_it = FALSE;
+
+    if ((config.sql_table_version < 2) && !assume_custom_table) {
+      if (config.what_to_count & COUNT_ID) {
+	Log(LOG_ERR, "ERROR: The use of IDs requires SQL table v2. Exiting.\n");
+        exit(1);	
+      }
+      else what_to_count ^= COUNT_ID;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "agent_id", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "agent_id=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_ID;
+      values[primitive].handler = where[primitive].handler = count_id_handler;
+      primitive++;
+    }
+  }
+
+  if (fakes & FAKE_SRC_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -564,7 +753,7 @@ int MY_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_MAC) {
+  if (fakes & FAKE_DST_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -578,7 +767,7 @@ int MY_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_HOST) {
+  if (fakes & FAKE_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -592,7 +781,7 @@ int MY_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_HOST) {
+  if (fakes & FAKE_DST_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -751,6 +940,7 @@ FILE *MY_file_open(const char *path, const char *mode)
     strlcpy(lh.sql_user, config.sql_user, DEF_HDR_FIELD_LEN);
     if (config.sql_host) strlcpy(lh.sql_host, config.sql_host, DEF_HDR_FIELD_LEN);
     else strlcpy(lh.sql_host, "localhost", DEF_HDR_FIELD_LEN);
+    lh.sql_table_version = config.sql_table_version;
     lh.sql_optimize_clauses = config.sql_optimize_clauses;
     lh.sql_history = config.sql_history;
     lh.what_to_count = config.what_to_count;
@@ -796,3 +986,16 @@ int MY_DB_Connect(struct DBdesc *db, char *host)
 
   return ret;
 }
+
+int MY_Exec(char *filename)
+{
+  int pid;
+
+  switch (pid = vfork()) {
+  case -1:
+    return -1;
+  case 0:
+    execv(filename, NULL); 
+    exit(0);
+  }
+}
