@@ -34,7 +34,7 @@
 #include "crc32.c"
 
 /* Functions */
-void pgsql_plugin(int pipe_fd, struct configuration *cfgptr) 
+void pgsql_plugin(int pipe_fd, struct configuration *cfgptr, void *ptr) 
 {
   struct pkt_data *data;
   struct networks_table nt;
@@ -48,8 +48,17 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
   struct pollfd pfd;
   struct insert_data idata;
   time_t t, now, refresh_deadline;
-  int timeout = config.sql_refresh_time*1000; /* dirty */ 
+  int timeout;
   int ret;
+#if defined (HAVE_MMAP)
+  struct ring *rg = &((struct channels_list_entry *)ptr)->rg;
+  struct ch_status *status = ((struct channels_list_entry *)ptr)->status;
+  u_int32_t bufsz = ((struct channels_list_entry *)ptr)->bufsize;
+
+  unsigned char *rgptr;
+  int pollagain = TRUE;
+  u_int32_t seq = 1, rg_err_count = 0;
+#endif
 
   /* XXX: glue */
   memcpy(&config, cfgptr, sizeof(struct configuration));
@@ -61,17 +70,23 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
   /* checks */
   if (config.what_to_count & COUNT_SUM_HOST) {
     Log(LOG_ERR, "ERROR: Option available only in memory table operations\nExiting ...\n\n");
-    exit(0);
+    exit(1);
   }
 
   if (!config.sql_refresh_time)
     config.sql_refresh_time = DEFAULT_DB_REFRESH_TIME;
 
+  if (!config.sql_table_version)
+    config.sql_table_version = DEFAULT_SQL_TABLE_VERSION;
+
+  timeout = config.sql_refresh_time*1000; /* dirty */
+
   if (config.networks_file) load_networks(config.networks_file, &nt);
   
   memset(sql_data, 0, sizeof(sql_data));
   memset(update_clause, 0, sizeof(update_clause));
   memset(insert_clause, 0, sizeof(insert_clause));
+  memset(lock_clause, 0, sizeof(lock_clause));
   memset(where, 0, sizeof(where));
   memset(values, 0, sizeof(values));
 
@@ -122,6 +137,28 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
     while ((t+idata.timeslot) < idata.basetime) t += idata.timeslot; 
     idata.basetime = t;
   }
+  
+  /* sql triggers time init: deadline; if a trigger exec is specified but no
+     time is supplied, use 'sql_refresh_time' as interval; this will result in
+     a trigger being executed each time data is purged into the DB */
+  if (config.sql_trigger_exec) {
+    time_t deadline;
+
+    deadline = now;
+    if (config.sql_trigger_time == COUNT_MINUTELY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*60;
+    else if (config.sql_trigger_time == COUNT_HOURLY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*3600;
+    else if (config.sql_trigger_time == COUNT_DAILY)
+      config.sql_trigger_time_howmany = config.sql_trigger_time_howmany*86400;
+    else config.sql_trigger_time_howmany = config.sql_refresh_time;
+
+    /* round off stuff */
+    t = roundoff_time(deadline);
+    while ((t+config.sql_trigger_time_howmany) < deadline) t += config.sql_trigger_time_howmany;
+    config.sql_trigger_time = t;
+    config.sql_trigger_time += config.sql_trigger_time_howmany; /* it's a deadline not a basetime */
+  }
 
   /* sql_refresh time init: deadline */
   refresh_deadline = now; 
@@ -141,8 +178,14 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
   memset(collision_queue, 0, cq_size*sizeof(struct db_cache *));
 
   /* plugin main loop */
-  while(1) {
+  for(;;) {
+    poll_again:
+#if defined (HAVE_MMAP)
+    status->wakeup = TRUE;
+#endif
     ret = poll(&pfd, 1, timeout);
+    if (ret < 0) goto poll_again;
+
     now = time(NULL);
 
     switch (ret) {
@@ -162,8 +205,12 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
 	  if (config.sql_backup_host) PG_compose_conn_string(&b, config.sql_backup_host);
 
 	  if (!PG_DB_Connect(&p)) Log(LOG_ALERT, "ALERT: PG_DB_Connect(): PGSQL daemon failed.\n");
-	  if (cq_ptr) PG_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
-          PG_cache_purge(queries_queue, qq_ptr, idata.num_primitives, TRUE);
+	  
+          PG_cache_purge(queries_queue, qq_ptr, idata.num_primitives, TRUE); 
+	  if (cq_ptr) {
+	    config.sql_dont_try_update = FALSE;
+	    PG_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE); 
+	  }
 
 	  if (!p.fail) PQfinish(p.desc);
 	  else if (b.connected) PQfinish(b.desc);
@@ -177,8 +224,47 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
       }
       break;
     default: /* poll(): received data */
-      if (read(pipe_fd, pipebuf, config.buffer_size) == 0) 
-        exit(0); /* we exit silently; something happened at the write end */
+#if !defined (HAVE_MMAP)
+      if ((ret = read(pipe_fd, pipebuf, config.buffer_size)) == 0) 
+        exit(1); /* we exit silently; something happened at the write end */
+
+      if (ret < 0) goto poll_again;
+#else
+      read_data:
+      if (!pollagain) {
+        seq++;
+        seq %= MAX_SEQNUM;
+        if (seq == 0) rg_err_count = FALSE;
+        now = time(NULL);
+      }
+      else {
+        if ((ret = read(pipe_fd, &rgptr, sizeof(rgptr))) == 0)
+          exit(1); /* we exit silently; something happened at the write end */
+      }
+
+      if (((struct ch_buf_hdr *)rg->ptr)->seq != seq) {
+        if (!pollagain) {
+          pollagain = TRUE;
+          goto poll_again;
+        }
+        else {
+          rg_err_count++;
+          if (config.debug || (rg_err_count > MAX_RG_COUNT_ERR)) {
+            Log(LOG_ERR, "ERROR: We are missing data.\n");
+            Log(LOG_ERR, "If you see this message once in a while, discard it. Otherwise some solutions follow:\n");
+            Log(LOG_ERR, "- increase shared memory size, 'plugin_pipe_size'; now: '%d'.\n", config.pipe_size);
+            Log(LOG_ERR, "- increase buffer size, 'plugin_buffer_size'; now: '%d'.\n", config.buffer_size);
+            Log(LOG_ERR, "- increase system maximum socket size.\n\n");
+          }
+          seq = ((struct ch_buf_hdr *)rg->ptr)->seq;
+        }
+      }
+
+      pollagain = FALSE;
+      memcpy(pipebuf, rg->ptr, bufsz);
+      if ((rg->ptr+bufsz) >= rg->end) rg->ptr = rg->base;
+      else rg->ptr += bufsz;
+#endif
 
       /* lazy sql refresh handling */ 
       if (now > refresh_deadline) {
@@ -197,8 +283,15 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
             if (config.sql_backup_host) PG_compose_conn_string(&b, config.sql_backup_host);
 
             if (!PG_DB_Connect(&p)) Log(LOG_ALERT, "ALERT: PG_DB_Connect(): PGSQL daemon failed.\n");
-	    if (cq_ptr) PG_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
             PG_cache_purge(queries_queue, qq_ptr, idata.num_primitives, TRUE);
+	    if (cq_ptr) {
+	      config.sql_dont_try_update = FALSE;
+	      PG_cache_purge(&collision_queue, cq_ptr, idata.num_primitives, FALSE);
+	    }
+
+	    if (config.sql_trigger_exec) {
+              if (now > config.sql_trigger_time) PG_Exec(config.sql_trigger_exec);
+            }
 
 	    if (!p.fail) PQfinish(p.desc);
 	    else if (b.connected) PQfinish(b.desc);
@@ -207,10 +300,19 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
 	    if (cq_ptr) cq_ptr = PG_cache_flush(&collision_queue, cq_ptr, FALSE);
             qq_ptr = PG_cache_flush(queries_queue, qq_ptr, TRUE);
 	    refresh_deadline += config.sql_refresh_time; 
+	    if (now > config.sql_trigger_time) config.sql_trigger_time  += config.sql_trigger_time_howmany;
             break;
 	  }
         }
       } 
+      else {
+        if (config.sql_trigger_exec) {
+          if (now > config.sql_trigger_time) {
+            PG_Exec(config.sql_trigger_exec);
+            config.sql_trigger_time  += config.sql_trigger_time_howmany;
+          }
+        }
+      }
 
       data = (struct pkt_data *) (pipebuf+sizeof(struct ch_buf_hdr));
       if (now > (idata.basetime + idata.timeslot)) idata.basetime += idata.timeslot;
@@ -223,10 +325,13 @@ void pgsql_plugin(int pipe_fd, struct configuration *cfgptr)
           binsearch(&nt, &data->primitives.dst_ip);
 
         idata.modulo = PG_cache_modulo(&data->primitives); 
-        PG_cache_insert(&data->primitives, data->pkt_len, &idata); 
+        PG_cache_insert(data, &idata); 
 	((struct ch_buf_hdr *)pipebuf)->num--;
 	if (((struct ch_buf_hdr *)pipebuf)->num) data++;
       }
+#if defined (HAVE_MMAP)
+      goto read_data;
+#endif
     }
   }
 }
@@ -235,7 +340,7 @@ unsigned int PG_cache_modulo(struct pkt_primitives *srcdst)
 {
   register unsigned int modulo;
 
-  modulo = cache_crc32((char *)srcdst, pp_size);
+  modulo = cache_crc32((unsigned char *)srcdst, pp_size);
   
   return modulo %= config.sql_cache_entries;
 }
@@ -243,20 +348,14 @@ unsigned int PG_cache_modulo(struct pkt_primitives *srcdst)
 int PG_cache_dbop(PGconn *db_desc, const struct db_cache *cache_elem, const int num_primitives)
 {
   PGresult *ret;
-  char *ptr_values, *ptr_where;
+  char *ptr_values, *ptr_where, *err_string;
   int num=0;
 
-  /* beginning DB transaction */
-  ret = PQexec(db_desc, "BEGIN");
-  if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
-    PQclear(ret);
-    return TRUE;
-  }
-  PQclear(ret);
-
   /* constructing SQL query */
   ptr_where = where_clause;
   ptr_values = values_clause; 
+  memset(where_clause, 0, sizeof(where_clause));
+  memset(values_clause, 0, sizeof(values_clause));
   while (num < num_primitives) {
     (*where[num].handler)(cache_elem, num, &ptr_values, &ptr_where);
     num++;
@@ -266,15 +365,19 @@ int PG_cache_dbop(PGconn *db_desc, const struct db_cache *cache_elem, const int
   strncat(sql_data, where_clause, SPACELEFT(sql_data));
 
   /* sending UPDATE query */
-  ret = PQexec(db_desc, sql_data);
-  if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
+  if (!config.sql_dont_try_update) {
+    ret = PQexec(db_desc, sql_data);
+    if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
+      err_string = PQresultErrorMessage(ret);
+      PQclear(ret);
+      Log(LOG_DEBUG, "FAILED query follows:\n%s\n", sql_data);
+      Log(LOG_ERR, "%s\n", err_string);
+      return TRUE;
+    }
     PQclear(ret);
-    Log(LOG_DEBUG, "%s\n\n", sql_data);
-    return TRUE;
   }
-  PQclear(ret);
 
-  if (!PG_affected_rows(ret)) {
+  if (config.sql_dont_try_update || (!PG_affected_rows(ret))) {
     /* UPDATE failed, trying with an INSERT query */ 
     strncpy(sql_data, insert_clause, sizeof(sql_data));
     snprintf(ptr_values, SPACELEFT(values_clause), ", %u, %u)", cache_elem->packet_counter, cache_elem->bytes_counter);
@@ -282,29 +385,29 @@ int PG_cache_dbop(PGconn *db_desc, const struct db_cache *cache_elem, const int
 
     ret = PQexec(db_desc, sql_data);
     if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
+      err_string = PQresultErrorMessage(ret);
       PQclear(ret);
-      Log(LOG_DEBUG, "%s\n\n", sql_data);
+      Log(LOG_DEBUG, "FAILED query follows:\n%s\n", sql_data);
+      Log(LOG_ERR, "%s\n", err_string);
+
       return TRUE;
     }
     PQclear(ret);
   }
 
-  /* Finishing DB transaction */
-  ret = PQexec(db_desc, "COMMIT");
-  if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
-    PQclear(ret);
-    return TRUE;
-  }
-
   if (config.debug) Log(LOG_DEBUG, "%s\n\n", sql_data);
 
   return FALSE;
 }
 
-void PG_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct insert_data *idata) 
+void PG_cache_insert(struct pkt_data *data, struct insert_data *idata)
 {
   int chances = 2;
   unsigned int modulo = idata->modulo;
+  unsigned long int basetime = idata->basetime;
+  struct pkt_primitives *srcdst = &data->primitives;
+
+  while ((basetime > data->pkt_time) && (data->pkt_time > 0)) basetime -= idata->timeslot;
 
   start:
   if (memcmp(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives)) != 0) { 
@@ -324,29 +427,15 @@ void PG_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct
           modulo--;
           modulo %= config.sql_cache_entries;
           idata->modulo = modulo;
-          chances--;
         }
-        else idata->modulo = config.sql_cache_entries;
+        else idata->modulo = config.sql_cache_entries-1;
+        chances--;
 
         goto start;
       }	
 
       if (config.debug) Log(LOG_DEBUG, "*** Entry aliasing ***\n");
-      if (cq_ptr < cq_size) {
-        memcpy(&collision_queue[cq_ptr], &cache[idata->modulo], dbc_size);
-        cq_ptr++;
-      }
-      else {
-        /* purging collision queue */
-        switch (fork()) {
-        case 0: /* Child */
-          PG_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
-          exit(0);
-        default: /* Parent */
-          cq_ptr = PG_cache_flush(&collision_queue, cq_ptr, FALSE);
-          break;
-        }
-      }
+      PG_handle_collision(&cache[idata->modulo]);
     }
     else {
       queries_queue[qq_ptr] = &cache[idata->modulo];
@@ -355,8 +444,8 @@ void PG_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct
 
     /* we add the new entry in the cache */
     memcpy(&cache[idata->modulo], srcdst, sizeof(struct pkt_primitives));
-    cache[idata->modulo].packet_counter = 1;
-    cache[idata->modulo].bytes_counter = ntohs(ip_len);
+    cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+    cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
     cache[idata->modulo].valid = TRUE;
     cache[idata->modulo].basetime = idata->basetime;
   }
@@ -365,35 +454,29 @@ void PG_cache_insert(struct pkt_primitives *srcdst, unsigned int ip_len, struct
       /* additional check: time */
       if ((cache[idata->modulo].basetime < idata->basetime) && (config.sql_history)) {
         if (config.debug) Log(LOG_DEBUG, "*** Entry aliasing ***\n");
-	if (cq_ptr < cq_size) {
-	  memcpy(&collision_queue[cq_ptr], &cache[idata->modulo], dbc_size);
-	  cq_ptr++;
-	}
-	else {
-	  /* purging collision queue */
-	  switch (fork()) {
-	  case 0: /* Child */
-	    PG_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
-	    exit(0);
-	  default: /* Parent */
-	    cq_ptr = PG_cache_flush(&collision_queue, cq_ptr, FALSE);
-	    break;
-	  }
-	}
-	cache[idata->modulo].packet_counter = 1;
-	cache[idata->modulo].bytes_counter = ntohs(ip_len);
+	PG_handle_collision(&cache[idata->modulo]);
+	cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+	cache[idata->modulo].bytes_counter =  ntohl(data->pkt_len);
 	cache[idata->modulo].basetime = idata->basetime;
       }
+      /* additional check: would counters overflow ? */
+      else if ((cache[idata->modulo].packet_counter > UINT32TMAX) ||
+	       (cache[idata->modulo].bytes_counter > UINT32TMAX)) {
+        PG_handle_collision(&cache[idata->modulo]);
+	cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+	cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
+        cache[idata->modulo].basetime = idata->basetime;
+      }
       else {
         /* everything is ok; summing counters */
-        cache[idata->modulo].packet_counter++;
-        cache[idata->modulo].bytes_counter += ntohs(ip_len);
+        cache[idata->modulo].packet_counter += ntohl(data->pkt_num);
+        cache[idata->modulo].bytes_counter += ntohl(data->pkt_len);
       }
     }
     else {
       /* entry invalidated by SIGALRM; restarting counters */
-      cache[idata->modulo].packet_counter = 1;
-      cache[idata->modulo].bytes_counter = ntohs(ip_len);
+      cache[idata->modulo].packet_counter = ntohl(data->pkt_num);
+      cache[idata->modulo].bytes_counter = ntohl(data->pkt_len);
       cache[idata->modulo].valid = TRUE;
       cache[idata->modulo].basetime = idata->basetime;
       queries_queue[qq_ptr] = &cache[idata->modulo];
@@ -406,23 +489,63 @@ void PG_cache_purge(struct db_cache *queue[], int index, const int num_primitive
 {
   unsigned char *qptr = (unsigned char *) *queue;
   struct logfile lf;
-  int j;
+  int j, r, reprocess = 0;
+  PGresult *ret;
 
   memset(&lf, 0, sizeof(struct logfile));
 
   if (config.debug) Log(LOG_DEBUG, "*** Purging cache - START ***\n");
 
-  if (ptr) for (j = 0; j < index; j++) PG_Query(&p, &b, &lf, queue[j], num_primitives);
+  /* beginning DB transaction */
+  ret = PQexec(p.desc, lock_clause);
+
+  /* for each element of the queue to be processed we execute PG_Query(); the function
+     returns a non-zero value if DB has failed; then, first failed element is saved to
+     allow reprocessing of previous elements if a failover method is in use; elements
+     need to be reprocessed because at the time of DB failure they were not yet committed */ 
+     
+  if (ptr) {
+    for (j = 0; j < index; j++) {
+      r = PG_Query(&p, &b, &lf, queue[j], num_primitives);
+      if ((!reprocess) && r) reprocess = j+1; /* avoding reprocess to be 0 when element j = 0 fails */
+    }
+  }
   else {
     for (j = 0; j < index; j++) {
-      PG_Query(&p, &b, &lf, (struct db_cache *) qptr, num_primitives);
+      r = PG_Query(&p, &b, &lf, (struct db_cache *) qptr, num_primitives);
+      if ((!reprocess) && r) reprocess = j+1;
       qptr += dbc_size;
     }
   }
 
+  /* Finishing DB transaction */
+  if (!p.fail) {
+    ret = PQexec(p.desc, "COMMIT");
+    if (PQresultStatus(ret) != PGRES_COMMAND_OK) {
+      p.fail = TRUE;
+      p.connected = FALSE;
+      reprocess = j+1;
+    }
+    PQclear(ret);
+  }
+
+  if (p.fail) {
+    reprocess--;
+    if (reprocess) {
+      qptr = (unsigned char *) *queue;
+      if (ptr) for (j = 0; j <= reprocess; j++) PG_Query(&p, &b, &lf, queue[j], num_primitives);
+      else {
+	for (j = 0; j <= reprocess; j++) {
+	  PG_Query(&p, &b, &lf, (struct db_cache *) qptr, num_primitives);
+	  qptr += dbc_size;
+	}
+      }
+    }
+  }
+
   /* rewinding stuff */
   if (lf.file) PG_file_close(&lf);
-  if ((lf.fail) || (b.fail)) Log(LOG_ALERT, "ALERT: recovery for PGSQL daemon failed.\n");
+  if ((lf.fail) || (b.fail)) Log(LOG_ALERT, "ALERT: recovery for PgSQL operation failed.\n");
   if (config.debug) Log(LOG_DEBUG, "*** Purging cache - END ***\n");
 }
 
@@ -444,6 +567,29 @@ int PG_cache_flush(struct db_cache *queue[], int index, int ptr)
   return index;
 }
 
+void PG_handle_collision(struct db_cache *elem)
+{
+  if (cq_ptr < cq_size) {
+    memcpy(&collision_queue[cq_ptr], elem, dbc_size);
+    cq_ptr++;
+  }
+  else {
+    /* purging collision queue */
+    switch (fork()) {
+    case 0: /* Child */
+      config.sql_dont_try_update = FALSE;
+      PG_cache_purge(&collision_queue, cq_ptr, num_primitives, FALSE);
+      exit(0);
+    default: /* Parent */
+      cq_ptr = PG_cache_flush(&collision_queue, cq_ptr, FALSE);
+      break;
+    }
+  
+    memcpy(&collision_queue[cq_ptr], elem, dbc_size);
+    cq_ptr++;
+  }
+}
+
 int PG_evaluate_history(int primitive)
 {
   if (config.sql_history) {
@@ -481,29 +627,34 @@ int PG_evaluate_history(int primitive)
 int PG_evaluate_primitives(int primitive)
 {
   register unsigned long int what_to_count=0;
+  register unsigned long int fakes=0;
+  short int assume_custom_table = FALSE;
 
-  if (config.sql_optimize_clauses) what_to_count = config.what_to_count;
+  if (config.sql_optimize_clauses) {
+    what_to_count = config.what_to_count;
+    assume_custom_table = TRUE;
+  }
   else {
     /* we are requested to avoid optimization; then we'll construct an
        all-true "what to count" bitmap. */ 
 
     if (config.what_to_count & COUNT_SRC_MAC) what_to_count |= COUNT_SRC_MAC;
-    else what_to_count |= FAKE_SRC_MAC;
+    else fakes |= FAKE_SRC_MAC;
     if (config.what_to_count & COUNT_DST_MAC) what_to_count |= COUNT_DST_MAC;
-    else what_to_count |= FAKE_DST_MAC;
+    else fakes |= FAKE_DST_MAC;
 
     if (config.what_to_count & COUNT_SRC_HOST) what_to_count |= COUNT_SRC_HOST;
     else {
-      if (frontend) what_to_count |= FAKE_SRC_HOST;
+      if (frontend) fakes |= FAKE_SRC_HOST;
       else what_to_count |= COUNT_SRC_HOST;
     }
     if (config.what_to_count & COUNT_DST_HOST) what_to_count |= COUNT_DST_HOST;
     else {
-      if (frontend) what_to_count |= FAKE_DST_HOST;
+      if (frontend) fakes |= FAKE_DST_HOST;
       else what_to_count |= COUNT_DST_HOST;
     }
 
-    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO;
+    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO|COUNT_ID|COUNT_VLAN;
   }
 
   /* 1st part: arranging pointers to an opaque structure and 
@@ -537,6 +688,33 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
+  if (what_to_count & COUNT_VLAN) {
+    int count_it = FALSE;
+
+    if ((config.sql_table_version < 2) && !assume_custom_table) {
+      if (config.what_to_count & COUNT_VLAN) {
+        Log(LOG_ERR, "ERROR: The use of VLAN accounting requires SQL table v2. Exiting.\n");
+        exit(1);
+      }
+      else what_to_count ^= COUNT_VLAN;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "vlan", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "vlan=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_VLAN;
+      values[primitive].handler = where[primitive].handler = count_vlan_handler;
+      primitive++;
+    }
+  }
+
   if (what_to_count & COUNT_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
@@ -625,7 +803,34 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_MAC) {
+  if (what_to_count & COUNT_ID) {
+    int count_it = FALSE;
+                                                                                                                                                 
+    if ((config.sql_table_version < 2) && !assume_custom_table) {
+      if (config.what_to_count & COUNT_ID) {
+        Log(LOG_ERR, "ERROR: The use of IDs requires SQL table v2. Exiting.\n");
+        exit(1);
+      }
+      else what_to_count ^= COUNT_ID;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "agent_id", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "agent_id=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_ID;
+      values[primitive].handler = where[primitive].handler = count_id_handler;
+      primitive++;
+    }
+  }
+
+  if (fakes & FAKE_SRC_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -639,7 +844,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_MAC) {
+  if (fakes & FAKE_DST_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -653,7 +858,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_HOST) {
+  if (fakes & FAKE_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -667,7 +872,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_HOST) {
+  if (fakes & FAKE_DST_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -696,6 +901,9 @@ int PG_compose_static_queries()
   primitives = PG_evaluate_primitives(primitives);
   strncat(insert_clause, ", packets, bytes)", SPACELEFT(insert_clause));
 
+  /* "LOCK ..." stuff */
+  snprintf(lock_clause, sizeof(lock_clause), "BEGIN; LOCK %s IN EXCLUSIVE MODE;", config.sql_table);
+
   /* "UPDATE ... SET ..." stuff */
   snprintf(update_clause, sizeof(update_clause), "UPDATE %s ", config.sql_table);
   strncat(update_clause, "SET packets=packets+%u, bytes=bytes+%u", SPACELEFT(update_clause));
@@ -747,13 +955,13 @@ void PG_exit_gracefully(int signum)
   exit(0); 
 }
 
-void PG_Query(struct DBdesc *p, struct DBdesc *b, struct logfile *lf, const struct db_cache *elem, int num_primitives)
+int PG_Query(struct DBdesc *p, struct DBdesc *b, struct logfile *lf, const struct db_cache *elem, int num_primitives)
 {
   PGresult *ret;
 
   if (!p->fail) {
     if (PG_cache_dbop(p->desc, elem, num_primitives)) goto recovery; 
-    else return;
+    else return FALSE;
   }
   else goto take_action;
 
@@ -787,6 +995,8 @@ void PG_Query(struct DBdesc *p, struct DBdesc *b, struct logfile *lf, const stru
       if (!lf->fail) fwrite(elem, sizeof(struct db_cache), 1, lf->file);
     }
   }
+
+  return TRUE;
 }
 
 void PG_file_close(struct logfile *lf)
@@ -811,6 +1021,7 @@ FILE *PG_file_open(const char *path, const char *mode)
     if (config.sql_host) strlcpy(lh.sql_host, config.sql_host, DEF_HDR_FIELD_LEN);
     else lh.sql_host[0] = '\0';
     lh.frontend = frontend;
+    lh.sql_table_version = config.sql_table_version;
     lh.sql_optimize_clauses = config.sql_optimize_clauses;
     lh.sql_history = config.sql_history;
     lh.what_to_count = config.what_to_count;
@@ -858,3 +1069,16 @@ static int PG_affected_rows(PGresult *result)
   return atoi(PQcmdTuples(result));
 }
 
+int PG_Exec(char *filename)
+{
+  int pid;
+
+  switch (pid = vfork()) {
+  case -1:
+    return -1;
+  case 0:
+    execv(filename, NULL);
+    exit(0);
+  }
+}
+
