@@ -105,9 +105,12 @@ int main(int argc, char **argv)
       do_nothing = TRUE;
       break;
     case 'h':
-    default:
       usage(argv[0]);
       exit(0);
+      break;
+    default:
+      usage(argv[0]);
+      exit(1);
     }
   }
 
@@ -192,7 +195,7 @@ int main(int argc, char **argv)
   PQfinish(p.desc);
   fclose(f);
 
-  return;
+  return 0;
 }
 
 int PG_cache_dbop(PGconn *db_desc, const struct db_cache *cache_elem, const int num_primitives)
@@ -292,29 +295,34 @@ int PG_evaluate_history(int primitive)
 int PG_evaluate_primitives(int primitive)
 {
   register unsigned long int what_to_count=0;
+  register unsigned long int fakes=0;
+  short int assume_custom_table = FALSE;
 
-  if (lh.sql_optimize_clauses) what_to_count = lh.what_to_count;
+  if (lh.sql_optimize_clauses) {
+    what_to_count = lh.what_to_count;
+    assume_custom_table = TRUE;
+  }
   else {
     /* we are requested to avoid optimization; then we'll construct an
        all-true "what to count" bitmap. */
 
     if (lh.what_to_count & COUNT_SRC_MAC) what_to_count |= COUNT_SRC_MAC;
-    else what_to_count |= FAKE_SRC_MAC;
+    else fakes |= FAKE_SRC_MAC;
     if (lh.what_to_count & COUNT_DST_MAC) what_to_count |= COUNT_DST_MAC;
-    else what_to_count |= FAKE_DST_MAC;
+    else fakes |= FAKE_DST_MAC;
 
     if (lh.what_to_count & COUNT_SRC_HOST) what_to_count |= COUNT_SRC_HOST;
     else {
-      if (lh.frontend) what_to_count |= FAKE_SRC_HOST;
+      if (lh.frontend) fakes |= FAKE_SRC_HOST;
       else what_to_count |= COUNT_SRC_HOST;
     }
     if (lh.what_to_count & COUNT_DST_HOST) what_to_count |= COUNT_DST_HOST;
     else {
-      if (lh.frontend) what_to_count |= FAKE_DST_HOST;
+      if (lh.frontend) fakes |= FAKE_DST_HOST;
       else what_to_count |= COUNT_DST_HOST;
     }
 
-    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO;
+    what_to_count |= COUNT_SRC_PORT|COUNT_DST_PORT|COUNT_IP_PROTO|COUNT_ID|COUNT_VLAN;
   }
 
   /* 1st part: arranging pointers to an opaque structure and 
@@ -348,6 +356,33 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
+  if (what_to_count & COUNT_VLAN) {
+    int count_it = FALSE;
+
+    if ((lh.sql_table_version < 2) && !assume_custom_table) {
+      if (lh.what_to_count & COUNT_VLAN) {
+        printf("ERROR: The use of VLAN accounting requires SQL table v2. Exiting.\n");
+        exit(1);
+      }
+      else what_to_count ^= COUNT_VLAN;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "vlan", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "vlan=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_VLAN;
+      values[primitive].handler = where[primitive].handler = count_vlan_handler;
+      primitive++;
+    }
+  }
+
   if (what_to_count & COUNT_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
@@ -436,7 +471,34 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_MAC) {
+  if (what_to_count & COUNT_ID) {
+    int count_it = FALSE;
+
+    if ((lh.sql_table_version < 2) && !assume_custom_table) {
+      if (lh.what_to_count & COUNT_ID) {
+        printf("ERROR: The use of IDs requires SQL table version 2. Exiting.\n");
+        exit(1);
+      }
+      else what_to_count ^= COUNT_ID;
+    }
+    else count_it = TRUE;
+
+    if (count_it) {
+      if (primitive) {
+        strncat(insert_clause, ", ", SPACELEFT(insert_clause));
+        strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
+        strncat(where[primitive].string, " AND ", sizeof(where[primitive].string));
+      }
+      strncat(insert_clause, "agent_id", SPACELEFT(insert_clause));
+      strncat(values[primitive].string, "%d", SPACELEFT(values[primitive].string));
+      strncat(where[primitive].string, "agent_id=%d", SPACELEFT(where[primitive].string));
+      values[primitive].type = where[primitive].type = COUNT_ID;
+      values[primitive].handler = where[primitive].handler = count_id_handler;
+      primitive++;
+    }
+  }
+
+  if (fakes & FAKE_SRC_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -450,7 +512,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_MAC) {
+  if (fakes & FAKE_DST_MAC) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -464,7 +526,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_SRC_HOST) {
+  if (fakes & FAKE_SRC_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
@@ -478,7 +540,7 @@ int PG_evaluate_primitives(int primitive)
     primitive++;
   }
 
-  if (what_to_count & FAKE_DST_HOST) {
+  if (fakes & FAKE_DST_HOST) {
     if (primitive) {
       strncat(insert_clause, ", ", SPACELEFT(insert_clause));
       strncat(values[primitive].string, ", ", sizeof(values[primitive].string));
