@@ -21,6 +21,7 @@
 
 /* prototypes */
 void setnonblocking(int);
+void setblocking(int);
 int daemonize();
 char *copy_argv(register char **);
 char *extract_token(char **, int);
@@ -28,6 +29,13 @@ char *extract_plugin_name(char **);
 void trim_spaces(char *);
 void trim_all_spaces(char *);
 int isblankline(char *);
+int iscomment(char *);
+int check_not_valid_char(char *, int);
 void debug_packet(struct pkt_data *);
 time_t roundoff_time(time_t);
 void write_pid_file(char *);
+int sanitize_buf_net(char *, int);
+int sanitize_buf(char *);
+void mark_columns(char *);
+int Setsocksize(int, int, int, void *, int);
+void *map_shared(void *, size_t, int, int, int, off_t);
